from __future__ import annotations

import shutil
import subprocess
import sys

from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING

from cleo.io.null_io import NullIO
from cleo.io.outputs.output import Verbosity
from poetry.core.constraints.version import Version
from poetry.core.constraints.version import parse_constraint

from poetry.utils._compat import decode
from poetry.utils.env.exceptions import NoCompatiblePythonVersionFoundError
from poetry.utils.env.script_strings import GET_PYTHON_VERSION_ONELINER


if TYPE_CHECKING:
    from cleo.io.io import IO

    from poetry.config.config import Config
    from poetry.poetry import Poetry


class Python:
    def __init__(self, executable: str | Path, version: Version | None = None) -> None:
        self.executable = Path(executable)
        self._version = version

    @property
    def version(self) -> Version:
        if not self._version:
            if self.executable == Path(sys.executable):
                python_version = ".".join(str(v) for v in sys.version_info[:3])
            else:
                encoding = "locale" if sys.version_info >= (3, 10) else None
                python_version = decode(
                    subprocess.check_output(
                        [str(self.executable), "-c", GET_PYTHON_VERSION_ONELINER],
                        text=True,
                        encoding=encoding,
                    ).strip()
                )
            self._version = Version.parse(python_version)

        return self._version

    @cached_property
    def patch_version(self) -> Version:
        return Version.from_parts(
            major=self.version.major,
            minor=self.version.minor,
            patch=self.version.patch,
        )

    @cached_property
    def minor_version(self) -> Version:
        return Version.from_parts(major=self.version.major, minor=self.version.minor)

    @staticmethod
    def _full_python_path(python: str) -> Path | None:
        # eg first find pythonXY.bat on windows.
        path_python = shutil.which(python)
        if path_python is None:
            return None

        try:
            encoding = "locale" if sys.version_info >= (3, 10) else None
            executable = subprocess.check_output(
                [path_python, "-c", "import sys; print(sys.executable)"],
                text=True,
                encoding=encoding,
            ).strip()
            return Path(executable)

        except subprocess.CalledProcessError:
            return None

    @staticmethod
    def _detect_active_python(io: IO) -> Path | None:
        io.write_error_line(
            "Trying to detect current active python executable as specified in"
            " the config.",
            verbosity=Verbosity.VERBOSE,
        )

        executable = Python._full_python_path("python")

        if executable is not None:
            io.write_error_line(f"Found: {executable}", verbosity=Verbosity.VERBOSE)
        else:
            io.write_error_line(
                "Unable to detect the current active python executable. Falling"
                " back to default.",
                verbosity=Verbosity.VERBOSE,
            )

        return executable

    @classmethod
    def get_system_python(cls) -> Python:
        return cls(executable=sys.executable)

    @classmethod
    def get_by_name(cls, python_name: str) -> Python | None:
        executable = cls._full_python_path(python_name)
        if not executable:
            return None

        return cls(executable=executable)

    @classmethod
    def get_preferred_python(cls, config: Config, io: IO | None = None) -> Python:
        io = io or NullIO()

        if not config.get("virtualenvs.use-poetry-python") and (
            active_python := Python._detect_active_python(io)
        ):
            return cls(executable=active_python)

        return cls.get_system_python()

    @classmethod
    def get_compatible_python(cls, poetry: Poetry, io: IO | None = None) -> Python:
        io = io or NullIO()
        supported_python = poetry.package.python_constraint
        python = None

        for suffix in [
            *sorted(
                poetry.package.AVAILABLE_PYTHONS,
                key=lambda v: (v.startswith("3"), -len(v), v),
                reverse=True,
            ),
            "",
        ]:
            if len(suffix) == 1:
                if not parse_constraint(f"^{suffix}.0").allows_any(supported_python):
                    continue
            elif suffix and not supported_python.allows_any(
                parse_constraint(suffix + ".*")
            ):
                continue

            python_name = f"python{suffix}"
            if io.is_debug():
                io.write_error_line(f"<debug>Trying {python_name}</debug>")

            executable = cls._full_python_path(python_name)
            if executable is None:
                continue

            candidate = cls(executable)
            if supported_python.allows(candidate.patch_version):
                python = candidate
                io.write_error_line(
                    f"Using <c1>{python_name}</c1> ({python.patch_version})"
                )
                break

        if not python:
            raise NoCompatiblePythonVersionFoundError(poetry.package.python_versions)

        return python



from __future__ import annotations

import os
import textwrap

from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from cleo.testers.application_tester import ApplicationTester

from poetry.console.application import Application
from poetry.console.commands.version import VersionCommand


if TYPE_CHECKING:
    from cleo.testers.command_tester import CommandTester
    from pytest_mock import MockerFixture

    from poetry.poetry import Poetry
    from tests.types import CommandTesterFactory
    from tests.types import FixtureDirGetter
    from tests.types import ProjectFactory


@pytest.fixture()
def command() -> VersionCommand:
    return VersionCommand()


@pytest.fixture
def tester(command_tester_factory: CommandTesterFactory) -> CommandTester:
    return command_tester_factory("version")


@pytest.fixture
def poetry_with_underscore(
    project_factory: ProjectFactory, fixture_dir: FixtureDirGetter
) -> Poetry:
    source = fixture_dir("simple_project")
    pyproject_content = (source / "pyproject.toml").read_text(encoding="utf-8")
    pyproject_content = pyproject_content.replace("simple-project", "simple_project")
    return project_factory(
        "project_with_underscore", pyproject_content=pyproject_content
    )


@pytest.mark.parametrize(
    "version, rule, expected",
    [
        ("0.0.0", "patch", "0.0.1"),
        ("0.0.0", "minor", "0.1.0"),
        ("0.0.0", "major", "1.0.0"),
        ("0.0", "major", "1.0"),
        ("0.0", "minor", "0.1"),
        ("0.0", "patch", "0.0.1"),
        ("1.2.3", "patch", "1.2.4"),
        ("1.2.3", "minor", "1.3.0"),
        ("1.2.3", "major", "2.0.0"),
        ("1.2.3", "prepatch", "1.2.4a0"),
        ("1.2.3", "preminor", "1.3.0a0"),
        ("1.2.3", "premajor", "2.0.0a0"),
        ("1.2.3-beta.1", "patch", "1.2.3"),
        ("1.2.3-beta.1", "minor", "1.3.0"),
        ("1.2.3-beta.1", "major", "2.0.0"),
        ("1.2.3-beta.1", "prerelease", "1.2.3b2"),
        ("1.2.3-beta1", "prerelease", "1.2.3b2"),
        ("1.2.3beta1", "prerelease", "1.2.3b2"),
        ("1.2.3b1", "prerelease", "1.2.3b2"),
        ("1.2.3", "prerelease", "1.2.4a0"),
        ("0.0.0", "1.2.3", "1.2.3"),
    ],
)
def test_increment_version(
    version: str, rule: str, expected: str, command: VersionCommand
) -> None:
    assert command.increment_version(version, rule).text == expected


@pytest.mark.parametrize(
    "version, rule, expected",
    [
        ("1.2.3", "prerelease", "1.2.4a0"),
        ("1.2.3a0", "prerelease", "1.2.3b0"),
        ("1.2.3a1", "prerelease", "1.2.3b0"),
        ("1.2.3b1", "prerelease", "1.2.3rc0"),
        ("1.2.3rc0", "prerelease", "1.2.3"),
        ("1.2.3-beta.1", "prerelease", "1.2.3rc0"),
        ("1.2.3-beta1", "prerelease", "1.2.3rc0"),
        ("1.2.3beta1", "prerelease", "1.2.3rc0"),
    ],
)
def test_next_phase_version(
    version: str, rule: str, expected: str, command: VersionCommand
) -> None:
    assert command.increment_version(version, rule, True).text == expected


def test_version_show(tester: CommandTester) -> None:
    tester.execute()
    assert tester.io.fetch_output() == "simple-project 1.2.3\n"


def test_version_show_with_underscore(
    command_tester_factory: CommandTesterFactory, poetry_with_underscore: Poetry
) -> None:
    tester = command_tester_factory("version", poetry=poetry_with_underscore)
    tester.execute()
    assert tester.io.fetch_output() == "simple_project 1.2.3\n"


def test_short_version_show(tester: CommandTester) -> None:
    tester.execute("--short")
    assert tester.io.fetch_output() == "1.2.3\n"


def test_version_update(tester: CommandTester) -> None:
    tester.execute("2.0.0")
    assert tester.io.fetch_output() == "Bumping version from 1.2.3 to 2.0.0\n"


def test_short_version_update(tester: CommandTester) -> None:
    tester.execute("--short 2.0.0")
    assert tester.io.fetch_output() == "2.0.0\n"


def test_phase_version_update(tester: CommandTester) -> None:
    assert isinstance(tester.command, VersionCommand)
    tester.command.poetry.package._set_version("1.2.4a0")
    tester.execute("prerelease --next-phase")
    assert tester.io.fetch_output() == "Bumping version from 1.2.4a0 to 1.2.4b0\n"


def test_dry_run(tester: CommandTester) -> None:
    assert isinstance(tester.command, VersionCommand)
    old_pyproject = tester.command.poetry.file.path.read_text(encoding="utf-8")
    tester.execute("--dry-run major")

    new_pyproject = tester.command.poetry.file.path.read_text(encoding="utf-8")
    assert tester.io.fetch_output() == "Bumping version from 1.2.3 to 2.0.0\n"
    assert old_pyproject == new_pyproject


def test_version_with_project_parameter(
    fixture_dir: FixtureDirGetter, mocker: MockerFixture
) -> None:
    app = Application()
    tester = ApplicationTester(app)

    orig_version_command = VersionCommand.handle

    def mock_handle(command: VersionCommand) -> int:
        exit_code = orig_version_command(command)

        command.io.write_line(f"ProjectPath: {command.poetry.pyproject_path.parent}")
        command.io.write_line(f"WorkingDirectory: {os.getcwd()}")

        return exit_code

    mocker.patch("poetry.console.commands.version.VersionCommand.handle", mock_handle)

    source_dir = fixture_dir("scripts")
    tester.execute(f"--project {source_dir} version")

    output = tester.io.fetch_output()
    expected = textwrap.dedent(f"""\
    scripts 0.1.0
    ProjectPath: {source_dir}
    WorkingDirectory: {os.getcwd()}
    """)

    assert source_dir != Path(os.getcwd())
    assert output == expected


def test_version_with_directory_parameter(
    fixture_dir: FixtureDirGetter, mocker: MockerFixture
) -> None:
    app = Application()
    tester = ApplicationTester(app)

    orig_version_command = VersionCommand.handle

    def mock_handle(command: VersionCommand) -> int:
        exit_code = orig_version_command(command)

        command.io.write_line(f"ProjectPath: {command.poetry.pyproject_path.parent}")
        command.io.write_line(f"WorkingDirectory: {os.getcwd()}")

        return exit_code

    mocker.patch("poetry.console.commands.version.VersionCommand.handle", mock_handle)

    source_dir = fixture_dir("scripts")
    tester.execute(f"--directory {source_dir} version")

    output = tester.io.fetch_output()
    expected = textwrap.dedent(f"""\
    scripts 0.1.0
    ProjectPath: {source_dir}
    WorkingDirectory: {source_dir}
    """)

    assert source_dir != Path(os.getcwd())
    assert output == expected


def test_version_with_directory_and_project_parameter(
    fixture_dir: FixtureDirGetter, mocker: MockerFixture
) -> None:
    app = Application()
    tester = ApplicationTester(app)

    orig_version_command = VersionCommand.handle

    def mock_handle(command: VersionCommand) -> int:
        exit_code = orig_version_command(command)

        command.io.write_line(f"ProjectPath: {command.poetry.pyproject_path.parent}")
        command.io.write_line(f"WorkingDirectory: {os.getcwd()}")

        return exit_code

    mocker.patch("poetry.console.commands.version.VersionCommand.handle", mock_handle)

    source_dir = fixture_dir("scripts")
    working_directory = source_dir.parent
    project_path = "./scripts"

    tester.execute(f"--directory {working_directory} --project {project_path} version")

    output = tester.io.fetch_output()

    expected = textwrap.dedent(f"""\
    scripts 0.1.0
    ProjectPath: {source_dir}
    WorkingDirectory: {working_directory}
    """)

    assert source_dir != working_directory
    assert output == expected



from __future__ import annotations

import os
import subprocess
import sys

from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from cleo.io.null_io import NullIO
from poetry.core.constraints.version import Version

from poetry.utils.env.python_manager import Python
from tests.utils.env.test_env_manager import check_output_wrapper


if TYPE_CHECKING:
    from pytest_mock import MockerFixture

    from poetry.config.config import Config
    from tests.types import ProjectFactory


def test_python_get_version_on_the_fly() -> None:
    python = Python(executable=sys.executable)

    assert python.executable == Path(sys.executable)
    assert python.version == Version.parse(
        ".".join([str(s) for s in sys.version_info[:3]])
    )
    assert python.patch_version == Version.parse(
        ".".join([str(s) for s in sys.version_info[:3]])
    )
    assert python.minor_version == Version.parse(
        ".".join([str(s) for s in sys.version_info[:2]])
    )


def test_python_get_system_python() -> None:
    python = Python.get_system_python()

    assert python.executable == Path(sys.executable)
    assert python.version == Version.parse(
        ".".join(str(v) for v in sys.version_info[:3])
    )


def test_python_get_preferred_default(config: Config) -> None:
    python = Python.get_preferred_python(config)

    assert python.executable == Path(sys.executable)
    assert python.version == Version.parse(
        ".".join(str(v) for v in sys.version_info[:3])
    )


def test_get_preferred_python_use_poetry_python_disabled(
    config: Config, mocker: MockerFixture
) -> None:
    mocker.patch(
        "subprocess.check_output",
        side_effect=check_output_wrapper(Version.parse("3.7.1")),
    )
    config.config["virtualenvs"]["use-poetry-python"] = False
    python = Python.get_preferred_python(config)

    assert python.executable.as_posix().startswith("/usr/bin/python")
    assert python.version == Version.parse("3.7.1")


def test_get_preferred_python_use_poetry_python_disabled_fallback(
    config: Config, mocker: MockerFixture
) -> None:
    config.config["virtualenvs"]["use-poetry-python"] = False
    mocker.patch(
        "subprocess.check_output",
        side_effect=subprocess.CalledProcessError(1, "some command"),
    )
    python = Python.get_preferred_python(config)

    assert python.executable == Path(sys.executable)


def test_fallback_on_detect_active_python(mocker: MockerFixture) -> None:
    m = mocker.patch(
        "subprocess.check_output",
        side_effect=subprocess.CalledProcessError(1, "some command"),
    )

    active_python = Python._detect_active_python(NullIO())

    assert active_python is None
    assert m.call_count == 1


@pytest.mark.skipif(sys.platform != "win32", reason="Windows only")
def test_detect_active_python_with_bat(tmp_path: Path) -> None:
    """On Windows pyenv uses batch files for python management."""
    python_wrapper = tmp_path / "python.bat"
    wrapped_python = Path(r"C:\SpecialPython\python.exe")
    encoding = "locale" if sys.version_info >= (3, 10) else None
    with python_wrapper.open("w", encoding=encoding) as f:
        f.write(f"@echo {wrapped_python}")
    os.environ["PATH"] = str(python_wrapper.parent) + os.pathsep + os.environ["PATH"]

    active_python = Python._detect_active_python(NullIO())

    assert active_python == wrapped_python


def test_python_find_compatible(project_factory: ProjectFactory) -> None:
    # Note: This test may fail on Windows systems using Python from the Microsoft Store,
    # as the executable is named `py.exe`, which is not currently recognized by
    # Python.get_compatible_python. This issue will be resolved in #2117.
    # However, this does not cause problems in our case because Poetry's own
    # Python interpreter is used before attempting to find another compatible version.
    fixture = Path(__file__).parent.parent / "fixtures" / "simple_project"
    poetry = project_factory("simple-project", source=fixture)
    python = Python.get_compatible_python(poetry)

    assert Version.from_parts(3, 4) <= python.version <= Version.from_parts(4, 0)
